(*  Title:      GCM.thy
    Author:     Ludovic Henrio 
                2008

    Note:      defines the structure of GCM components, based on the notions of
                 - futures, requests
                 - interfaces, and bindings
                 - hierarchical composition
                 - valueas are abstracted away by naturals, but track of future references is kept
                 - behaviour can be specified by primitive components, 
                      composite behaviour is entirely defined by the semantics
*)

header {* GCM Component *}

theory GCM imports FMap Lemmas 
begin

subsection{* First defines Names *}
(* for simplicity Names are isomorphic to a finite subset of nats *)
consts max_name :: nat

typedef Name = "{n :: nat. n <= max_name }"
by auto;

lemma finite_Name_set: "finite {n :: nat. n <= max_name }";
apply (rule_tac n = "max_name + 1" and f = "% x. x" in nat_seg_image_imp_finite);
apply (simp add: ran_def);
apply force;
done;

lemma Univ_abs_name: "(UNIV :: (Name set)) = Abs_Name ` {n :: nat. n <= max_name }";
apply (subgoal_tac "! x :: Name. x : Abs_Name ` {n :: nat. n <= max_name }");
apply force;
apply (rule allI);
apply (subgoal_tac "Abs_Name(Rep_Name x) = x");
prefer 2;
apply (rule Rep_Name_inverse) ;
apply (erule subst);
apply (rule imageI);
apply (fold Name_def);
apply (rule Rep_Name);
done;

lemma finite_Name: "finite (UNIV :: (Name set))";
apply (subst Univ_abs_name);
apply (rule finite_imageI);
apply (rule finite_Name_set);
done;

instance Name :: "fintype" 
proof 
show "finite (UNIV :: (Name set))" by (rule finite_Name);
qed;

types       Fid = "nat\<times>Name" (*types for future ids =  id + name of the cp computing the result *)

types Value = "nat\<times>(Fid list)"
 (* a nat symbolising the result + the list of  futures referenced by the value *)


datatype ItfKind= Client | Server

datatype Cardinality = Single

(* interface definition, types could be added here *)
record Interface = 
  kind::ItfKind
  cardinality::Cardinality


datatype QualifiedName = This Name 
                         | Sub Name  Name (infix ".\<^sub>i" 50)

subsection{*Component and Requests structure *}
record Binding = 
  src:: QualifiedName 
  dest:: QualifiedName


record Request= 
id::Fid
parameter:: Value 
invokedItf:: Name

record Result= 
fid::Fid
fValue:: Value


datatype LCstate = Started | Stopped

(* behaviour of primitive components *)
datatype Action = Call Name Value Fid (*called interface + parameter + created id *)
                | ReceiveResult Fid Value (* future id x value received *)
  (*                | Wait Fid   future id --apparently not useful for the moment ... *)
                | EndService Fid Value (* value of the result for the request *)
                | NewService Name Value Fid (* name of the invoked interface, value of the parameter, corresponding future *)
                | Tau 

record intState = (* Primitive internal state *)
PNFstate:: LCstate
PRqRefs:: "Fid list" (* the list of request ids referenced by the state *)
Pstate:: nat (* a representation of the internal state: nat for simple *)
PcurrentReqs:: "Fid list" (*"Fid option" *)
         (* id of the current requests *)

constdefs reachable :: "(intState \<times> Action \<times> intState) set \<Rightarrow> intState set " 
          "reachable beh == {s'.\<exists> s a . (s,a,s')\<in>beh} "

typedef  Behaviours="{beh:: (intState \<times> Action \<times> intState) set. 
   (\<forall> s s' . ((s,Tau,s')\<in> beh \<longrightarrow> (set (PRqRefs s')\<subseteq>set (PRqRefs s))
                        \<and> (PcurrentReqs s') = (PcurrentReqs s))) \<and> (* tau operations can only forget futures *)
   (\<forall> s s' i v f. ((s,NewService i v f,s')\<in> beh \<longrightarrow> 
                        (\<forall> v' f'. \<exists> s' . (s,NewService i v' f',s')\<in>beh) \<and>
                        set (PRqRefs s')\<subseteq>set (PRqRefs s)\<union>set (snd v) \<and>
                        (PcurrentReqs s') = f#(PcurrentReqs s) ))\<and> 
                       (*\<and>  add served itf belong to the component in correctcomponent structure *)   
    (\<forall> s s' i v f.((s,Call i v f,s')\<in>beh \<longrightarrow>( (* Call operations imply: *)
                        (\<forall> f'. \<exists> s'' . (s,Call i v f',s'')\<in>beh) (*any future id can be created*)
                        \<and>set (snd v)\<subseteq>set (PRqRefs s)                 (* sent futures included in known futures *)
                        \<and>   set (PRqRefs s')\<subseteq> (set (PRqRefs s)\<union>{f})
                        \<and> (PcurrentReqs s') = (PcurrentReqs s) ))) \<and>   (* only new future is f *)
                       (*\<and>  add invoked itf belong to the component in correctcomponent structure *)   
   (\<forall> s\<in>reachable beh .\<forall> f. (\<forall> v. \<exists> s' . (s,ReceiveResult f v,s')\<in>beh)) \<and> 
                                                (* it is always possible to receive a future value for a  future *)
   (\<forall> s s' v f.((s,ReceiveResult f v,s')\<in>beh \<longrightarrow> (               (* ReceiveFuture implies: *)
                        set (PRqRefs s')\<subseteq> (set (PRqRefs s)-{f})\<union> set(snd v)
                        \<and> (PcurrentReqs s') = (PcurrentReqs s) 
                        \<and> (f\<notin>set(PRqRefs s) \<longrightarrow>s=s') ))) \<and>
   (\<forall> s s' v f.((s,EndService f v,s')\<in>beh \<longrightarrow> (          (* EndService implies: *) 
                        (set(PRqRefs s')\<subseteq> set(PRqRefs s)) (* EndService operations can only forget futures *)
                        \<and> (set(snd v) \<subseteq> set(PRqRefs s))  (* sent value only contain known futures *)
                        \<and> f\<in>set (PcurrentReqs s)
                          \<and> (PcurrentReqs s') = remove1 f (PcurrentReqs s) )))        
                   }"
apply (unfold reachable_def);
by auto;

typedef WaitingStateFunction="{wait:: intState => Fid list. \<forall> s f. (f\<in> set (wait s)  \<longrightarrow>f\<in>set (PRqRefs s)) }"
by auto;

record PrimState = 
Pqueue:: "Request list" (* pending*)
PintState:: intState
PcomputedResults:: "Result list" (* could be a set *)
behaviour:: Behaviours (* the local semantics/behaviour of the component *)
                                                    (* the only infinite set *)
WaitingState:: WaitingStateFunction
PfutrMap::"Fid-~>Name list"

definition Behaviour :: "PrimState \<Rightarrow> (intState\<times>Action\<times>intState) set"
where "Behaviour s \<equiv> Rep_Behaviours(behaviour s)"

definition ApplyWait :: "PrimState \<Rightarrow> intState \<Rightarrow> Fid list"
where "ApplyWait s \<equiv> (Rep_WaitingStateFunction(WaitingState s))"

record CompState = 
Cqueue:: "Request list" (* pending*)
CNFstate:: LCstate
CcomputedResults:: "Result list"
CfutrMap::"Fid-~>Name list"

datatype Component =
             (* interfaces *)    (*sub components *) (* bindings: finite set *) (* state *)
  Primitive  Name "Name -~> Interface"                               PrimState |
  Composite  Name "Name -~> Interface" "Component list" "Binding set" CompState


primrec getName:: "Component\<Rightarrow>Name" where
  "getName (Primitive N itf s) = N" |
  "getName (Composite N itf sub b s) = N"

primrec getItfs:: "Component\<Rightarrow>Name -~> Interface" where
  "getItfs (Primitive N itf s) = itf" |
  "getItfs (Composite N itf sub b s) = itf"


primrec getCp:: "Component list \<Rightarrow> Name \<Rightarrow> Component option" (infix "^" 60) where
  "getCp [] N' = None" |
  "getCp (C#CL) N' = (if (getName C=N') then Some C else (CL^N') )"

primrec getSubCp:: "Component \<Rightarrow> Name \<Rightarrow> Component option" where
  "getSubCp (Primitive N itf s) N' = None" |
  "getSubCp (Composite N itf sub b s) N' = sub^N"

primrec changeSubCp:: "Component \<Rightarrow>Component\<Rightarrow> Component" (infix "<<-" 60) where 
              (*replaces the subcomponent, does nothing if there is no subcomponent with this name*)
  "changeSubCp (Primitive N itf s)  C'= (Primitive N itf s)" |
  "changeSubCp (Composite N itf sub b s)  C'= Composite N itf (map (\<lambda> C. if getName C=(getName C') then C' else C) sub) b s"

primrec changeCp:: "Component list \<Rightarrow>Component\<Rightarrow> Component list" (infix "<-" 60) where 
              (*replaces the subcomponent, does nothing if there is no subcomponent with this name*)
  "changeCp [] C = []" |
  "changeCp (C#CL) C' = (if (getName C=getName C') then (C'#CL) else C#(CL<-C') )"

primrec removeSubCp:: "Component \<Rightarrow> Name \<Rightarrow> Component" where 
              (*removes the subcomponent, does nothing if there is no subcomponent with this name*)
  "removeSubCp (Primitive N itf s) N' = (Primitive N itf s)" |
  "removeSubCp (Composite N itf sub b s) N' = (case sub^N' of
            None => (Composite N itf sub b s) |
            Some C => Composite N itf (remove1 C sub) b s)"

primrec addSubCp:: "Component \<Rightarrow> Component\<Rightarrow> Component" where 
              (*adds the subcomponent, or replaces it if there is already a cp with the same name*)
  "addSubCp (Primitive N itf s)  C'= (Primitive N itf s)" |
  "addSubCp (Composite N itf sub b s)  C'= (case sub^(getName C') of
            None \<Rightarrow> (Composite N itf (C'#sub) b s) |
            Some C \<Rightarrow>(Composite N itf sub b s)<<-C')"

inductive
 equiv:: "[Component, Component] \<Rightarrow> bool" (infix "\<approx>" 55) where
  "(Primitive N itf s)\<approx> (Primitive N itf s) "    |
   "   length sub = length sub' \<and> (\<forall> C\<in>set sub. \<exists> C'\<in>set sub'. C\<approx>C')
             \<Longrightarrow>(Composite N itf sub b s)\<approx>(Composite N itf sub' b s)"

subsection{*Various basic functions*}
 constdefs Src_belongs :: "QualifiedName \<Rightarrow> Binding set \<Rightarrow> bool" (infix "\<in>\<^sub>b" 60) 
    "Src_belongs n b == n\<in> (\<lambda>x. src x) ` b"

(* to be changed if multicast interfaces ! *)
  constdefs NoDuplicateSrc:: "Binding set \<Rightarrow> bool" 
      "NoDuplicateSrc bindings==
        \<forall> b1\<in> bindings. \<forall> b2\<in>bindings. b1\<noteq>b2\<longrightarrow>src b1\<noteq>src b2" 

primrec SymmetricalKind:: "ItfKind \<Rightarrow>ItfKind" where
  "SymmetricalKind Server = Client " |
  "SymmetricalKind Client = Server"

constdefs ExternalToInternal:: "Interface\<Rightarrow>Interface" 
"ExternalToInternal i == \<lparr>kind = SymmetricalKind (kind i),cardinality = (cardinality i)\<rparr>"


primrec GetQualified:: "QualifiedName \<Rightarrow> Component \<Rightarrow> Interface option" where
  "GetQualified qn (Primitive N itfs s) = None" |
  "GetQualified qn (Composite N itfs subCp bindings s) = 
  (case qn of 
      (This n) \<Rightarrow>  (Option.map ExternalToInternal (itfs n)) |
      (CN.\<^sub>iIN) \<Rightarrow> (case (subCp ^ CN) of
         None \<Rightarrow> None |
         Some Cp \<Rightarrow> ((getItfs Cp) IN)))"


(* cpList returns the list of components taking part in a composition  order not deterministic !*)
primrec cpList:: "Component \<Rightarrow> Component list"  and
 cpListlist:: "Component list \<Rightarrow> Component list" 
 where
  "cpList (Primitive N itfs s) = [(Primitive N itfs s)]" |
  "cpList (Composite N itfs subCp bindings s) =  (Composite N itfs subCp bindings s)# (cpListlist subCp)" | 
  "cpListlist [] = []" |
  "cpListlist (C#CL) = (cpList C)@ cpListlist CL"

constdefs cpSet:: "Component \<Rightarrow> Component set"
"cpSet C == set (cpList C)"

constdefs cpListset:: "Component list \<Rightarrow> Component set"
"cpListset CL == set (cpListlist CL)"

primrec getRecSubCp:: "Component \<Rightarrow> Name \<Rightarrow> Component option" (infix "^^" 60) and
getRecSubCpList:: "Component list \<Rightarrow> Name \<Rightarrow> Component option"
 where
  "getRecSubCp (Primitive N itf s) N' = (if (N=N') then Some (Primitive N itf s) else None)" |
  "getRecSubCp (Composite N itf sub b s) N' = (if (N=N') then Some (Composite N itf sub b s) 
                                                        else (getRecSubCpList sub N'))" |
  "getRecSubCpList [] N = None " |
  "getRecSubCpList (C#CL) N= (case (getRecSubCp C N) of
             Some C' \<Rightarrow> Some C' |
             None \<Rightarrow> (getRecSubCpList CL N))"



(* rqidlist returens the list of request Ids in all components odf a composition order not deterministic !*)
primrec RqIdList:: "Component \<Rightarrow> Fid list" and
        RqIdListList:: "Component list \<Rightarrow> Fid list" where
  "RqIdList (Primitive N itfs s) = (map (\<lambda>r. fid r) (PcomputedResults s))
                                @ (PcurrentReqs (PintState s))
                                @(map (\<lambda>r. id r) (Pqueue s))"                |
  "RqIdList (Composite N itfs subCp bindings s) =(map (\<lambda>r. fid r) (CcomputedResults s))
                                                  @(map (\<lambda>r. id r) (Cqueue s)) @
                                                 (RqIdListList subCp) " |
  "RqIdListList [] = []" |
  "RqIdListList (C#CL)=(RqIdList C)@RqIdListList (CL)"


(* ReferencedRqs returens the set of request Ids referenced in all components*)
primrec ReferencedRqs::  "Component \<Rightarrow> Fid set" and
        ReferencedRqsList:: "Component list \<Rightarrow> Fid set" where
  "ReferencedRqs (Primitive N itfs s) = (foldl (\<lambda>x y. x\<union> set (snd(fValue y)))
                                      (foldl (\<lambda>x y. x\<union> set (snd(parameter y))) (set (PRqRefs  (PintState s))) (Pqueue s)) (PcomputedResults s))" |
  "ReferencedRqs (Composite N itfs subCp bindings s) = (foldl (\<lambda>x y. x\<union> set (snd (fValue y)))
                                   (foldl (\<lambda>x y. x\<union> set (snd (parameter y))) {} (Cqueue s)) (CcomputedResults s)) \<union>
                                   (ReferencedRqsList subCp) " |
  "ReferencedRqsList [] = {}" |
  "ReferencedRqsList (C#CL)=(ReferencedRqs C)\<union>ReferencedRqsList (CL)"

(* ComputedRqSet returens the list of request Ids in all components of a composition order not deterministic !*)
primrec ComputedRqs::  "Component \<Rightarrow> Fid set" and
        ComputedRqsList:: "Component list \<Rightarrow> Fid set" where
  "ComputedRqs (Primitive N itfs s) = (set (map fid (PcomputedResults s)))" |
  "ComputedRqs (Composite N itfs subCp bindings s) =  (set (map fid (CcomputedResults s))) \<union>
                                                 (ComputedRqsList subCp) " |
  "ComputedRqsList [] = {}" |
  "ComputedRqsList (C#CL)=(ComputedRqs C)\<union>ComputedRqsList (CL)"
 

primrec getQueue:: " Component \<Rightarrow> Request list" where 
 "getQueue (Primitive N itfs s)=(Pqueue s)" |
 "getQueue (Composite N itfs subCp bindings s)=(Cqueue s)"

primrec getNFstate:: " Component \<Rightarrow> LCstate" where 
 "getNFstate (Primitive N itfs s)=(PNFstate (PintState s))" |
 "getNFstate (Composite N itfs subCp bindings s)=(CNFstate s)"

primrec getComputedResults:: " Component \<Rightarrow> Result list" where 
 "getComputedResults (Primitive N itfs s)=(PcomputedResults s)" |
 "getComputedResults (Composite N itfs subCp bindings s)=(CcomputedResults s)"


primrec findInResultList:: "Result list \<Rightarrow> Fid \<Rightarrow> Value option" where
 "findInResultList [] n = None" |
 "findInResultList (R#RL) n = (if (fid R = n) then Some (fValue R) else findInResultList RL n)"


primrec findResult:: " Component \<Rightarrow> Fid \<Rightarrow> Value option"
  and   findResultList:: " Component list \<Rightarrow> Fid \<Rightarrow> Value option" where 
 "findResult (Primitive N itfs s) n= (findInResultList (PcomputedResults s) n)" |
 "findResult (Composite N itfs subCp bindings s) n=
     (case (findInResultList (CcomputedResults s) n) of
	Some x\<Rightarrow>Some x |
	None \<Rightarrow> findResultList subCp n)" |
 "findResultList [] f= None" |
 "findResultList (C#CL) f=(case (findResult C f) of
             Some x \<Rightarrow> Some x |
             None \<Rightarrow>(findResultList CL f))"


primrec enQueue:: "Component \<Rightarrow> Request \<Rightarrow> Component" (infix "\<leftarrow>" 70) where
 "enQueue (Primitive N itfs s) R = (Primitive N itfs (s\<lparr>Pqueue:=(Pqueue s) @ [R]\<rparr>))" |
 "enQueue (Composite N itfs subCp bindings s) R = (Composite N itfs subCp bindings (s\<lparr>Cqueue:=(Cqueue s) @ [R]\<rparr>))"

subsection{*define correct components *}
(* only considers single interfaces *)
(* each request is targetted at a defined itf
   each binding has a single source.
   each binding corresponds to existing source and dest
   and recursively for each component
  *) 
primrec CorrectComponentStructure :: "Component \<Rightarrow> bool" 
and CorrectComponentStructureList:: "Component list \<Rightarrow> bool"  where
  "CorrectComponentStructure (Primitive N itfs s) = 
      ((\<forall>Q\<in>set (Pqueue s). (invokedItf Q)\<in> dom itfs \<and> kind (the (itfs (invokedItf Q))) = Server)
      \<and> (\<forall> s1 N v f s2. ((s1,Call N v f,s2) \<in> Behaviour s\<longrightarrow>  N\<in> dom itfs \<and> kind (the (itfs N)) = Client))
      \<and> (\<forall> s1 N v f s2. ((s1,NewService N v f,s2) \<in> Behaviour s\<longrightarrow>  N\<in> dom itfs \<and> kind (the (itfs N)) = Server)))" |
  "CorrectComponentStructure (Composite N itfs subCp bindings cs) = ((\<forall> b\<in> bindings. 
     (GetQualified (src b) (Composite N itfs subCp bindings cs) =Some\<lparr>kind=Client,cardinality=Single\<rparr>)
      \<and> (GetQualified (dest b) (Composite N itfs subCp bindings cs)=Some\<lparr>kind=Server,cardinality=Single\<rparr>))
     \<and> NoDuplicateSrc bindings
     \<and>  distinct (map getName subCp)
     \<and> (\<forall>Q\<in>set (Cqueue cs). (invokedItf Q)\<in> dom itfs \<and> kind (the (itfs (invokedItf Q))) = Server)
     \<and> CorrectComponentStructureList subCp)" |
  " CorrectComponentStructureList [] = True" |
  " CorrectComponentStructureList (C#CL) = (CorrectComponentStructure C \<and> CorrectComponentStructureList CL)" 

(* a cp is correct if each request id is unique 
   and each referenced request exists
   and the component structure is correct *)
constdefs CorrectComponent:: "Component\<Rightarrow>bool" 
"CorrectComponent c == CorrectComponentStructure c \<and> distinct (RqIdList c) \<and> (ReferencedRqs c) \<subseteq>(set (RqIdList c)) \<and> distinct (map getName (cpList c)) \<and> (\<forall> f\<in>set (RqIdList c). snd f \<in>set (map getName (cpList c)))"

(* much weaker than correctcomponent does not check correct referencing, compose hierarchically *)
constdefs CorrectComponentWeak:: "Component \<Rightarrow>bool"  
"CorrectComponentWeak c == CorrectComponentStructure c \<and> distinct (RqIdList c) 
                      \<and> distinct (map getName (cpList c)) "
constdefs CorrectComponentWeakList:: "Component list\<Rightarrow>bool"  
"CorrectComponentWeakList CL == (CorrectComponentStructureList CL) 
                       \<and> distinct (RqIdListList CL) 
                       \<and> distinct (map getName (cpListlist CL)) "


subsection{* additional constructs (mainly useful for proofs) *}

primrec RqIdLocalList:: "Component \<Rightarrow> Fid list"  where
  "RqIdLocalList (Primitive N itfs s) = (map (\<lambda>r. fid r) (PcomputedResults s))
                                @ (PcurrentReqs (PintState s))
                                @(map (\<lambda>r. id r) (Pqueue s))"                |
  "RqIdLocalList (Composite N itfs subCp bindings s) =(map (\<lambda>r. fid r) (CcomputedResults s))
                                                  @(map (\<lambda>r. id r) (Cqueue s)) " 


primrec getFutrMap:: "Component \<Rightarrow> Fid-~>Name list" where
  "getFutrMap (Primitive N itfs s) = (PfutrMap s)"                |
  "getFutrMap (Composite N itfs subCp bindings s) = CfutrMap s" 


primrec setFutrMap:: "Component \<Rightarrow> Fid-~>Name list \<Rightarrow>Component" where
  "setFutrMap (Primitive N itfs s) newFM= (Primitive N itfs (s\<lparr>PfutrMap:=newFM\<rparr>))"               |
  "setFutrMap (Composite N itfs subCp bindings s) newFM= (Composite N itfs subCp bindings (s\<lparr>CfutrMap:=newFM\<rparr>))" 


lemma get_set_futrMap[simp]:
 "getFutrMap(setFutrMap C FM) = FM"
by (case_tac C,auto);

lemma set_get_futrMap[simp]:
 "setFutrMap C (getFutrMap C) = C"
by (case_tac C,auto);


lemma SymmetricalKind_Server[simp]: "(SymmetricalKind k = Server) = (k=Client)"
by (case_tac k, auto)
lemma SymmetricalKind_Client[simp]: "(SymmetricalKind k = Client) = (k=Server)"
by (case_tac k, auto)

(* an induction principle for components *)
lemma induct_Comp:
"\<lbrakk> \<And> N cplist itfs bindings s . P (Primitive N  itfs s);
 \<And> N cplist itfs bindings s . (\<forall> C\<in> set cplist. P C)\<Longrightarrow> P (Composite N  itfs cplist bindings s)\<rbrakk>\<Longrightarrow> P Comp"
apply (subgoal_tac "\<forall>C CL .  (P C \<and>  (foldl (op \<and>)  True (map P CL)))");
apply simp;
apply (rule allI)+;
apply (induct_tac C and CL);
apply auto;
apply (rotate_tac 1);
apply (drule_tac x=Name in meta_spec);
apply (drule_tac x=list in meta_spec);
apply (drule_tac x=fun1 in meta_spec);
apply (drule_tac x=fun2 in meta_spec);
apply (drule_tac x=CompState_ext_type in meta_spec);
apply (erule meta_impE);
apply auto;
apply (subgoal_tac " \<forall> x. ((foldl op \<and> True (map P list) \<and> x\<in>set list) \<longrightarrow>P x)");
apply simp;
apply (induct_tac list)
apply force;
apply clarify;
apply (simp);
apply auto;
apply (insert foldl_and);
apply (drule_tac x="P a" in meta_spec);
apply (drule_tac x="(map P lista)" in meta_spec);
apply simp+;
apply (drule_tac x=xa in spec);
apply simp+;
apply (insert foldl_and_rec);
apply auto;
done;


end;
